/*-------------------------------------------------------------------------
 *
 * receiver_raw.c
 *		Receive and apply logical changes generated by decoder_raw. This
 *		creates some basics for a multi-master cluster using vanilla
 *		PostgreSQL without modifying its code.
 *
 * Copyright (c) 1996-2016, PostgreSQL Global Development Group
 *
 * IDENTIFICATION
 *		receiver_raw/receiver_raw.c
 *
 *-------------------------------------------------------------------------
 */

/* Some general headers for custom bgworker facility */
#include "postgres.h"
#include "main.h"
#include "gstore_shmem.h"
#include "fmgr.h"
#include "libpq-fe.h"
#include "pqexpbuffer.h"
#include "replication/logical.h"
#include "replication/output_plugin.h"
#include "replication/slot.h"
#include "access/xact.h"
#include "lib/stringinfo.h"
#include "pgstat.h"
#include "executor/spi.h"
#include "postmaster/bgworker.h"
#include "storage/ipc.h"
#include "storage/latch.h"
#include "storage/proc.h"
#include "utils/guc.h"
#include "utils/snapmgr.h"
#include "utils/dsa.h"
#include <unistd.h>

static long receiverTime = 0;
/* Signal handling */
static volatile sig_atomic_t got_sigterm = false;
static volatile sig_atomic_t got_sighup = false;

/* GUC variables */
static char *receiver_database = "postgres";
static char *receiver_slot = "slot";
static char *receiver_conn_string = "replication=database dbname=postgres";
static int receiver_idle_time = 100;
static bool receiver_sync_mode = true;

/* Worker name */
static char *worker_name = "bgworker";

/* Stream functions */
static void fe_sendint64(int64 i, char *buf);
static int64 fe_recvint64(char *buf);

/* Backgroundworker Main function */
void receiver_raw_main(Datum main_arg);

static char *plugin = "graph_index";
// static char *plugin = "test_decoding";

dsa_area *proc_dsa_area = NULL;

static void
receiver_raw_sigterm(SIGNAL_ARGS)
{
	int save_errno = errno;
	got_sigterm = true;
	if (MyProc)
		SetLatch(&MyProc->procLatch);
	errno = save_errno;
}

static void
receiver_raw_sighup(SIGNAL_ARGS)
{
	int save_errno = errno;
	got_sighup = true;
	if (MyProc)
		SetLatch(&MyProc->procLatch);
	errno = save_errno;
}


void receiver_raw_main(Datum main_arg)
{
	elog(LOG, "Inside receiver_raw_main");
	GraphIndexDSAProcStateStruct *bgworker_dsa_local;
	/* Register functions for SIGTERM/SIGHUP management */
	pqsignal(SIGHUP, receiver_raw_sighup);
	pqsignal(SIGTERM, receiver_raw_sigterm);

	/* We're now ready to receive signals */
	BackgroundWorkerUnblockSignals();
	
	/* Get DSA Local Pointer */
	bgworker_dsa_local = get_dsa_proc_state();
	proc_dsa_area = bgworker_dsa_local->area;
	elog(LOG, "proc_dsa_area addr ==> %p", proc_dsa_area);

	dsa_pointer XidDMLData;
	XidDMLDataList *XidDMLData_local;
	char *name;
	char *val = "world";
	// sleep(60);
	elog(LOG, "WAL after sleep before while");
	int i = 0;
	/* sample allocations using DSA */
	while(i < 10){
		XidDMLData = dsa_allocate(proc_dsa_area, sizeof(XidDMLDataList));
		XidDMLData_local = (XidDMLDataList *)dsa_get_address(proc_dsa_area, XidDMLData);
		XidDMLData_local->xid = 'h';
		XidDMLData_local->name = dsa_allocate(proc_dsa_area, sizeof(char)*10);
		name = (char *)dsa_get_address(proc_dsa_area, XidDMLData_local->name);
		memcpy(name, val, strlen(val) + 1);
		i++;
		elog(LOG, "i = %d", i);
		dsa_dump(proc_dsa_area);
		sleep(10);
	}
	// while(true){}
	/* No problems, so clean exit */
	proc_exit(0);
}

/*
 * Entry point to load parameters
 */
static void
receiver_raw_load_params(void)
{
	/*
	 * Defines database where to connect and apply the changes.
	 */
	DefineCustomStringVariable("receiver_raw.database",
							   "Database where changes are applied.",
							   "Default value is \"postgres\".",
							   &receiver_database,
							   "postgres",
							   PGC_POSTMASTER,
							   0, NULL, NULL, NULL);

	/* Slot used to get replication changes on remote */
	DefineCustomStringVariable("receiver_raw.slot_name",
							   "Replication slot used for logical changes.",
							   "Default value is \"slot\".",
							   &receiver_slot,
							   "slot",
							   PGC_POSTMASTER,
							   0, NULL, NULL, NULL);

	/*
	 * Connection string used to connect to remote source.
	 * Note: This could be made far better as if user does not set
	 * replication START_REPLICATION will simply fail :)
	 */
	DefineCustomStringVariable("receiver_raw.conn_string",
							   "Replication slot used for logical changes.",
							   "Default value is \"slot\".",
							   &receiver_conn_string,
							   "replication=database dbname=postgres",
							   PGC_POSTMASTER,
							   0, NULL, NULL, NULL);

	/* Nap time between two loops */
	DefineCustomIntVariable("receiver_raw.idle_time",
							"Nap time between two successive loops (ms).",
							"Default value set to 100 ms.",
							&receiver_idle_time,
							100, 1, 10000,
							PGC_SIGHUP,
							0, NULL, NULL, NULL);

	/* Synchronous mode */
    DefineCustomBoolVariable("receiver_raw.sync_mode",
							 "Enforce feedback to server.",
							 NULL,
							 &receiver_sync_mode,
							 true,
							 PGC_SIGHUP,
							 0, NULL, NULL, NULL);
}

/*
 * Entry point for worker loading
 */
void
receiver_initialization(void)
{
	BackgroundWorker worker;

	receiver_raw_load_params();

	/* Worker parameter and registration */
	MemSet(&worker, 0, sizeof(BackgroundWorker));
	worker.bgw_flags = BGWORKER_SHMEM_ACCESS |
		BGWORKER_BACKEND_DATABASE_CONNECTION;
	worker.bgw_start_time = BgWorkerStart_ConsistentState;
	snprintf(worker.bgw_library_name, sizeof(worker.bgw_library_name),
		"graph_index");
	snprintf(worker.bgw_function_name, sizeof(worker.bgw_function_name),
		"receiver_raw_main");
	// worker.bgw_main = receiver_raw_main;
	snprintf(worker.bgw_name, BGW_MAXLEN, "%s", worker_name);
	/* Wait 10 seconds for restart before crash */
	worker.bgw_restart_time = 10;
	worker.bgw_main_arg = (Datum) 0;
	worker.bgw_notify_pid = 0;
	RegisterBackgroundWorker(&worker);
}
